#include "internet.h"

#undef printf

#define printf(format, args...)

#define DETECT_PPPOE "/tmp/detect_pppoe"
#define DETECT_DHCP "/tmp/detect_dhcp"
#define DETECT_PPTP "/tmp/detect_pptp"

enum {
	IF_NONE = 0,
	IF_UP,
	IF_DOWN
};

int ifconfig(char *ifname, int flags, char *addr, char *mask)
{
	int fd, val;
	struct ifreq ifr;
	struct sockaddr_in *sin;

	fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (fd < 0)
		return 0;

	val = 0;
	strncpy(ifr.ifr_name, ifname, IFNAMSIZ);

	if (flags) {
		if (ioctl(fd, SIOCGIFFLAGS, &ifr) < 0)
			goto ret;	
		if (flags == IF_UP)
			ifr.ifr_flags |= IFF_UP |IFF_RUNNING;
		else	/* IF_DOWN */
			ifr.ifr_flags &= ~IFF_UP;
		if (ioctl(fd, SIOCSIFFLAGS, &ifr) < 0)
			goto ret;
	}

	sin = (struct sockaddr_in *)(&ifr.ifr_addr); // union ...

	if (addr) {
		sin->sin_family = AF_INET;
		inet_aton(addr, &sin->sin_addr);
		if (ioctl(fd, SIOCSIFADDR, &ifr) < 0)
			goto ret;
	}

	if (mask) {
		sin->sin_family = AF_INET;
		inet_aton(mask, &sin->sin_addr);
                if (ioctl(fd, SIOCSIFNETMASK, &ifr) < 0)
			goto ret;
        }

	val = 1;

ret:
	close(fd);
	return val;
}

/*======================================================
                                                    [ PPPoE Detection ]
    ======================================================*/

/* an NUL 'Service-Name' tag && a 'Host-Uniq' tag with PID value */
#define PADI_PAYLOAD_SIZE	((2 * PPPOE_TAG_SIZE) + 8)

static int init_PADI(struct pppoe_packet *packet, uint8 *hwaddr)
{
	struct pppoe_tag *svc;
	char id[8];

	memset(packet->h_dest, 0xFF, ETH_ALEN);
	memcpy(packet->h_source, hwaddr, ETH_ALEN);
	packet->h_proto = htons(ETH_PPPOE_DISC);

	packet->ver_type = 0x11;
	packet->code = CODE_PADI;
	packet->sid = 0;
	packet->length = htons(PADI_PAYLOAD_SIZE);

	/* Add an NUL 'Service-Name' tag */
	svc = (struct pppoe_tag *)(packet->data);
	svc->tag_type = htons(TAG_SERVICE_NAME);
	svc->tag_len = 0;

	/* Add 'Host-Uniq' tag with PID value */
	svc++;
	svc->tag_type = htons(TAG_HOST_UNIQ);
	svc->tag_len = htons(8);
	memcpy(svc->tag_data, id, 8);

	return (ETH_HDRSIZE + PPPOE_HDRSIZE + PADI_PAYLOAD_SIZE);
}

static int open_DiscSocket(uint8 *hwaddr)
{
	int fd;
	int optval = 1;    	
	struct ifreq ifr;
	struct sockaddr_ll sa;

	memset(&sa, 0, sizeof(sa));
	fd = socket(PF_PACKET, SOCK_RAW, htons(ETH_PPPOE_DISC));
	if (fd < 0) {
		printf("PPPoE socket failed.\n");
		return -1;
	}
	
	if (setsockopt(fd, SOL_SOCKET, SO_BROADCAST, &optval, sizeof(optval)) < 0)
		goto err;
	
	strncpy(ifr.ifr_name, NET_IFNAME, sizeof(ifr.ifr_name));
	if (ioctl(fd, SIOCGIFHWADDR, &ifr) < 0) {
		printf("Can't get " NET_IFNAME " mac.\n");
		goto err;
	}
	memcpy(hwaddr, ifr.ifr_hwaddr.sa_data, ETH_ALEN);
	
	strncpy(ifr.ifr_name, NET_IFNAME, sizeof(ifr.ifr_name));
	if (ioctl(fd, SIOCGIFINDEX, &ifr) < 0)
		goto err;	

	sa.sll_ifindex = ifr.ifr_ifindex;
	sa.sll_family = AF_PACKET;
	sa.sll_protocol = htons(ETH_PPPOE_DISC);
	if (bind(fd, (struct sockaddr *)&sa, sizeof(sa)) < 0) {
		printf("PPPoE bind failed.\n");
		goto err;
	}

	return fd;
	
err:
	close(fd);
	return -1;
}

static void parsePADOPacket(struct pppoe_packet *packet, 
		int *acName, int *svcName)
{
	uint16 len;
	uint8 *curTag;
	uint8 *tagData;
	uint16 tagType, tagLen;
		
	len = ntohs(packet->length);
	if (len > PPPOE_DATA_LEN) {
		printf("Invalid PPPoE packet length (%u)", len);
		return;
	}
	
	if (packet->ver_type != 0x11) {
		printf("Invalid version or type.\n");
		return;
	}
	
	curTag = packet->data;
	while (curTag - packet->data < len) {
		tagType	= TAG_TYPE(curTag);
		tagLen	= TAG_LEN(curTag);
		tagData	= curTag + PPPOE_TAG_SIZE;
		
		if (tagType == TAG_END_OF_LIST)
			return;
		
		if ((tagData - packet->data) + tagLen > len) {
			printf("Invalid PPPoE tag length (%u)", tagLen);
			return;
		}

		if (tagType == TAG_AC_NAME) {
			*acName = 1;
			printf("Access-Concentrator: %.*s\n", (int) tagLen, tagData);
		} else if (tagType == TAG_SERVICE_NAME) {
			*svcName = 1;
			printf("Service-Name: %.*s\n", (int) tagLen, tagData);
		}

		curTag = tagData + tagLen;
	}
}

static int waitForPADO(int fd, uint8 *hwaddr)
{
	int r, len;
	int tries = 5;
	int acName;
	int svcName;
	fd_set readable;
	struct timeval tv;
	struct pppoe_packet packet;
	
	for (;;) {
		if (--tries < 0)
			return 0;

		tv.tv_sec = PPPOE_TIMEOUT;
		tv.tv_usec = 0;
		FD_ZERO(&readable);
		FD_SET(fd, &readable);
		
		r = select(fd + 1, &readable, NULL, NULL, &tv);
		if (r < 0)
			continue;
		if (r == 0)
			return 0;
		len = recv(fd, &packet, sizeof(struct pppoe_packet), 0);
		if (ntohs(packet.length) + ETH_HDRSIZE + PPPOE_HDRSIZE > len ||
		    memcmp(packet.h_dest, hwaddr, ETH_ALEN) ||
		    packet.code != CODE_PADO)
			continue;

		acName = svcName = 0;
		parsePADOPacket(&packet, &acName, &svcName);
		if (acName && svcName)
			return 1;
	}

	return 0;
}

#if 0
static int PPPoE_Detection(void)
{
	int i;
	int ret, fd, plen;
	uint8 hwaddr[ETH_ALEN];
	struct pppoe_packet padi;

	ret = 0;
	fd = open_DiscSocket(hwaddr);
	if (fd == -1)
		return 0;

	plen = init_PADI(&padi, hwaddr);

	for (i = 0; i < 3; i++) {
		send(fd, &padi, plen, 0);
		
		ret = waitForPADO(fd, hwaddr);
		if (ret == 1)
			break;
	}

	close(fd);
	return ret;
}
#else
static int PPPoE_Detection(void)
{
	char value[8] = "";
	FILE *fp;

	unlink(DETECT_PPPOE);
	system("/usr/sbin/detect pppoe");
	sleep(3);
	if ((fp = fopen(DETECT_PPPOE, "r"))) {
		fgets(value, sizeof(value), fp);
		fclose(fp);

		if (value && *value == '1'){
			system("rm /tmp/detect_pppoe");
			return 1;
		}
		system("rm /tmp/detect_pppoe");
	}

	return 0;
}
#endif

/*======================================================
                                                  [ DHCP & BPA Detection ]
    ======================================================*/

/* Compute Internet Checksum for "count" bytes
  * beginning at location "addr".
 */
static uint16 calc_csum(void *addr, int count)
{
	int sum = 0;
	uint16 *source = (uint16 *)addr;

	while (count > 1) {
		sum += *source++;
		count -= 2;
	}

	/*  add left-over byte, if any */
	if (count > 0) {
		/* Make sure that the left-over byte is added correctly both
		 * with little and big endian hosts */
		uint16 left = 0;
		*(uint8 *) (&left) = *(uint8 *)source;
		sum += left;
	}

	/*  fold 32-bit sum to 16 bits */
	while (sum >> 16)
		sum = (sum & 0xFFFF) + (sum >> 16);

	return (uint16)(~sum);
}

static int read_interface(char *ifname, int *ifindex, uint8 *arp)
{
	int fd, ret;
	struct ifreq ifr;

	fd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
	if (fd < 0)
		return 0;

	ret = 1;
	memset(&ifr, 0, sizeof(struct ifreq));
	ifr.ifr_addr.sa_family = AF_INET;
	strcpy(ifr.ifr_name, ifname);

	if (ioctl(fd, SIOCGIFINDEX, &ifr) == 0)
		*ifindex = ifr.ifr_ifindex;
	else
		ret = 0;

	if (ioctl(fd, SIOCGIFHWADDR, &ifr) == 0)
		memcpy(arp, ifr.ifr_hwaddr.sa_data, 6);
	else
		ret = 0;

	close(fd);
	return ret;
}

static int raw_socket(int ifindex)
{
	int fd;
	struct sockaddr_ll addr;

	fd = socket(PF_PACKET, SOCK_DGRAM, htons(ETH_P_IP));
	if (fd < 0)
		return -1;

	memset(&addr, 0, sizeof(addr));
	addr.sll_family = AF_PACKET;
	addr.sll_protocol = htons(ETH_P_IP);
	addr.sll_ifindex = ifindex;
	if (bind(fd, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
		close(fd);
		return -1;
	}

	return fd;
}

static int add_option(uint8 *op, uint8 code, int len, void *data)
{	
	op[OPT_CODE] = code;
	op[OPT_LEN] = (uint8)len;

	memcpy(&op[OPT_DATA], data, len);

	return (len + 2);
}

static uint8 *get_option(struct dhcpMessage *packet, uint8 opcode, uint8 *oplen)
{
	int i, len;
	uint8 *op;
	int over = 0, done = 0, curr = OPTION_FIELD;

	i = 0;
	len = OPT_SIZE;
	op = packet->options;

	while(!done) {
		if (op[i + OPT_CODE] == opcode) {
			if (i + 1 + op[i + OPT_LEN] >= len)
				return NULL;
			*oplen = op[i + OPT_LEN];
			return op + i + 2;
		}

		switch (op[i + OPT_CODE]) {
		case DHCP_PADDING:
			i++;
			break;

		case DHCP_OPTION_OVER:
			if (i + 1 + op[i + OPT_LEN] >= len)
				return NULL;

			over = op[i + 3];
			i += op[OPT_LEN] + 2;
			break;

		case DHCP_END:
			if (curr == OPTION_FIELD && over & FILE_FIELD) {
				op = packet->file;
				i = 0;
				len = 128;
				curr = FILE_FIELD;
			} else if (curr == FILE_FIELD && over & SNAME_FIELD) {
				op = packet->sname;
				i = 0;
				len = 64;
				curr = SNAME_FIELD;
			} else {
				done = 1;
			}
			break;

		default:
			i += op[OPT_LEN + i] + 2;
		}
	}

	return NULL;
}

static int init_discover(struct rawPacket *packet, uint32 xid, uint8 *arp)
{
	int dhcp_len;
	int packet_len;
	uint8 msgtype, *op, id[7];
	struct dhcpMessage *dmsg;
	static uint8 reqlist[] = { 0x01, 0x03, 0x06, 0x0c, 0x0f, 0x1c };

	msgtype = DHCPDISCOVER;
	id[0] = ETH_10MB;
	memcpy(&id[1], arp, 6);
	memset(packet, 0, sizeof(struct rawPacket));

	/*********************************************************/
	dmsg = &packet->data;
	dmsg->op = BOOTREQUEST;
	dmsg->htype = ETH_10MB;
	dmsg->hlen = ETH_10MB_LEN;
	dmsg->xid = htonl(xid);
	dmsg->cookie = htonl(DHCP_MAGIC);

	memcpy(dmsg->chaddr, arp, 6);

	/* option */
	op = dmsg->options;
	op += add_option(op, DHCP_MESSAGE_TYPE, 1, &msgtype);
	op += add_option(op, DHCP_CLIENT_ID, 7, id);
	op += add_option(op, DHCP_VENDOR, sizeof(VENDOR) - 1, VENDOR);
	op += add_option(op, DHCP_PARAM_REQ, sizeof(reqlist), reqlist);
	*op = DHCP_END;

	dhcp_len = sizeof(struct dhcpMessage) -sizeof(dmsg->options) + 3 +
				(op - dmsg->options); /* two bytes padding */

	/*********************************************************/
	packet_len = sizeof(struct udphdr) + dhcp_len;
	packet->udp.source = htons(CLIENT_PORT);
	packet->udp.dest = htons(SERVER_PORT);
	packet->udp.len = htons(packet_len);
	packet->udp.check = 0; /* zero-csum */

	/*********************************************************/
	packet_len += sizeof(struct iphdr);
	packet->ip.version = 4;
	packet->ip.ihl = 5; /* sizeof(struct iphdr) >> 2 */
	packet->ip.tot_len = htons(packet_len);
	packet->ip.ttl = IPDEFTTL;
	packet->ip.protocol = IPPROTO_UDP;
	packet->ip.daddr = 0xFFFFFFFF;
	packet->ip.check = calc_csum(&(packet->ip), sizeof(packet->ip));

	return packet_len;
}

/* Is it a DHCP-OFFER packet ? */
static int verify_packet(int fd, uint32 xid, uint8 *arp, int *bpain)
{
	int bytes;
	uint8 *opdata, oplen;
	char domain[OPT_SIZE];
	struct rawPacket packet;
	struct dhcpMessage *dhcp;

	bytes = read(fd, &packet, sizeof(struct rawPacket));
	if (bytes < PKTSIZE_NO_OPT ||bytes < ntohs(packet.ip.tot_len))
		return 0;

	if (!UDP_IP(&packet.ip) ||packet.udp.dest != htons(CLIENT_PORT))
		return 0;

	dhcp = &packet.data;
	if (ntohl(dhcp->cookie) != DHCP_MAGIC)
		return 0;
	if (dhcp->xid != htonl(xid) ||memcmp(dhcp->chaddr, arp, 6))
		return 0;

	opdata = get_option(dhcp, DHCP_MESSAGE_TYPE, &oplen);
	if (opdata == NULL ||*opdata != DHCPOFFER)
		return 0;
	opdata = get_option(dhcp, DHCP_DOMAIN_NAME, &oplen);
	if (opdata) {
		sprintf(domain, "%.*s", (int)oplen, opdata);
		printf("DHCP-Domain : %s\n", domain);
		if (strstr(domain, BPA_SVR_SUFFIX))
			*bpain = 1;
	}

	return 1;
}

static int wait_dhcp_offer(int fd, uint32 xid, uint8 *arp, int *bpain)
{
	int r;
	long timeo;
	fd_set readable;
	struct timeval tv;

	timeo = uptime() + DHCP_TIMEOUT;

	for (;;) {
		FD_ZERO(&readable);
		FD_SET(fd, &readable);

		tv.tv_sec = timeo - uptime();
		if (tv.tv_sec <= 0)
			return 0;
		tv.tv_usec = 0;

		r = select(fd + 1, &readable, NULL, NULL, &tv);
		if (r < 1)
			return 0;
		if (verify_packet(fd, xid, arp, bpain))
			return 1;
	}

	return 0;
}

static int DHCP_Detection(int *bpain)
{
	char value[8] = "";
	FILE *fp;

	unlink(DETECT_DHCP);
	system("/usr/sbin/detect dhcp");
	sleep(3);
	if ((fp = fopen(DETECT_DHCP, "r"))) {
		fgets(value, sizeof(value), fp);
		fclose(fp);

		if (value && *value == '1'){
			system("rm /tmp/detect_dhcp");
			return 1;
		}
		system("rm /tmp/detect_dhcp");
	}

	return 0;
}

/*======================================================
                                                    [ PPTP Detection ]
    ======================================================*/

static int open_pptpsock(struct in_addr inetaddr) 
{
	int s;
	struct timeval timo;
	struct sockaddr_in dest;

	memset(&dest, 0, sizeof(struct sockaddr_in));
	dest.sin_family = AF_INET;
	dest.sin_port   = htons(PPTP_PORT);
	dest.sin_addr   = inetaddr;

	s = socket(AF_INET, SOCK_STREAM, 0);
	if (s < 0)
		return -1;

	timo.tv_sec = PPTP_TIMEOUT;
	timo.tv_usec = 0;
	setsockopt(s, SOL_SOCKET, SO_SNDTIMEO, &timo, sizeof(timo));
	setsockopt(s, SOL_SOCKET, SO_RCVTIMEO, &timo, sizeof(timo));
	if (connect(s, (struct sockaddr *) &dest, sizeof(dest)) < 0) {
		close(s); 
		return -1;
	}

	/* Make this socket non-blocking. */
	fcntl(s, F_SETFL, O_NONBLOCK);

	return s;
}

static int wait_ctrl_reply(int fd)
{
	int len;
	fd_set readable;
	struct timeval timo;
	char packet[MAX_PKT_SIZE];
	struct pptp_start_ctrl_conn *repl;

	timo.tv_sec = PPTP_TIMEOUT;
	timo.tv_usec = 0;
	FD_ZERO(&readable);
	FD_SET(fd, &readable);
	if (select(fd + 1, &readable, NULL, NULL, &timo) < 1)
		return 0;

	len = recv(fd, packet, sizeof(packet), 0);
	if (len < sizeof(struct pptp_start_ctrl_conn))
		return 0;

	repl = (struct pptp_start_ctrl_conn *)packet;
	if (repl->header.magic != htonl(PPTP_MAGIC) ||
	    repl->header.pptp_type != htons(PPTP_MESSAGE_CONTROL) ||
	    repl->header.ctrl_type != htons(PPTP_START_CTRL_CONN_RPLY))
		return 0;

	/* This MAY BE a real PPTP server ^_^ */
	return 1;
}

static void init_ctrl_rqst(struct pptp_start_ctrl_conn *req)
{
	memset(req, 0, sizeof(*req));

	req->header.length	= htons(sizeof(*req));
	req->header.pptp_type	= htons(PPTP_MESSAGE_CONTROL);
	req->header.magic		= htonl(PPTP_MAGIC);
	req->header.ctrl_type	= htons(PPTP_START_CTRL_CONN_RQST);

	req->version		= htons(PPTP_VERSION);
	req->framing_cap	= htonl(PPTP_FRAME_CAP);
	req->bearer_cap	= htonl(PPTP_BEARER_CAP);
	req->max_channels= htons(PPTP_MAX_CHANNELS);
	req->firmware_rev	= htons(PPTP_FIRMWARE_VERSION);

	strcpy((char *)req->hostname, PPTP_HOSTNAME);
	strcpy((char *)req->vendor, PPTP_VENDOR);
}

static int PPTP_Detection(void)
{

	char value[8] = "";
	FILE *fp;

	unlink(DETECT_PPTP);
	system("/usr/sbin/detect pptp");
	sleep(3);
	if ((fp = fopen(DETECT_PPTP, "r"))) {
		fgets(value, sizeof(value), fp);
		fclose(fp);

		if (value && *value == '1'){
			system("rm /tmp/detect_pptp");
			return 1;
		}
		system("rm /tmp/detect_pptp");
	}

	return 0;
}

/*======================================================
                                         [ Detection WAN Connection Type ]
    ======================================================*/

int Internet_Detection(char *type)
{
	int dhcp, bpa;

	system("sbin/cmdwan.sh stop");

	if (!ifconfig(NET_IFNAME, IF_UP, NULL, NULL)) {
		printf("INET_DET: can't bring "NET_IFNAME "up!\n");
		return 0;
	}

	dhcp = bpa = 0;

	if (dhcp = DHCP_Detection(NULL))
		strcpy(type, "DHCP");
	else if (PPPoE_Detection())
		strcpy(type, "PPPoE");
	/*else if ((dhcp = DHCP_Detection(&bpa)) && bpa)
		strcpy(type, "BigPond");*/
	else if (PPTP_Detection())
		strcpy(type, "PPTP");
	else
		strcpy(type, "Static");

	ifconfig(NET_IFNAME, IF_DOWN, NULL, NULL);

	system("/sbin/cmdwan.sh restart");

	return 1;
}

